import pytest

from pagidantic import PagidanticDict, PagidanticSet, Paginator, pagidantic


def dict_to_list(dict_object: dict) -> list[dict]:
    """Transform dict to list of dicts."""
    return [{k: v} for k, v in dict_object.items()]


@pytest.fixture()
def list_of_int() -> list[int]:
    """Generate list of integers."""
    return list(range(1, 51))


@pytest.fixture()
def list_of_dict() -> list[dict]:
    """Generate list of dictionary."""
    d = {f"{x}": x for x in range(1, 21)}
    return dict_to_list(d)


@pytest.fixture()
def dict_data() -> dict:
    """Generate single dictionary."""
    return {f"{x}": x for x in range(1, 11)}


@pytest.fixture()
def paginator(list_of_int) -> Paginator:
    """Return paginator object with list of integers as input."""
    return pagidantic(object_list=list_of_int)


@pytest.fixture()
def second_paginator(list_of_dict) -> Paginator:
    """Return paginator object with list of dictionary as input, start page set to 1, limit 3."""
    return pagidantic(list_of_dict, page_limit=3, start_page=1)


@pytest.fixture()
def paginator_dict_proxy(dict_data) -> PagidanticDict:
    """Return PaginatorDictProxy object with dictionary as input."""
    return PagidanticDict(dict_data, page_limit=5)


@pytest.fixture()
def paginator_set_proxy() -> PagidanticSet:
    """Return PaginatorSetProxy object with set as input, limit set to 3."""
    return PagidanticSet({1, 2, 3, 4, 5, 6, 7, 8}, page_limit=3)



from pagidantic.paginator import Paginator


class PagidanticDict(Paginator):
    """
    Initialize a Paginator for paginating a dictionary.

    :param object_list: The dictionary to be paginated.
    :type object_list: dict
    :param page_limit: Number of items per page.
    :type page_limit: int, optional
    :param start_page: The page to start pagination from.
    :type start_page: int, optional
    """

    def __init__(self, object_list: dict, page_limit: int = 10, start_page: int = 0):
        self.object_list = object_list
        super().__init__(
            object_list=self._dict_to_list(),
            page_limit=page_limit,
            start_page=start_page,
        )

    def _dict_to_list(self) -> list[dict]:
        """Transform dict to list of dicts."""
        if not isinstance(self.object_list, dict):
            raise TypeError(f"Expected dict object, not {type(self.object_list)}")
        return [{k: v} for k, v in self.object_list.items()]



import pytest
from conftest import dict_to_list

from pagidantic import PagidanticDict, PagidanticSet


class TestPaginatorDictProxy:
    """Tests for Test PagidanticDict"""

    def test_init(self, paginator_dict_proxy: PagidanticDict, dict_data: dict):
        """Validate init assigment."""
        assert paginator_dict_proxy.object_list == dict_to_list(dict_data)
        assert paginator_dict_proxy.page_limit == 5
        assert paginator_dict_proxy.start_page == 0

    def test_invalid_object_list(self):
        """Validate object_list param."""
        with pytest.raises(TypeError):
            PagidanticDict([1, 2, 3])
        with pytest.raises(TypeError):
            PagidanticDict({1, 2, 3})
        with pytest.raises(TypeError):
            PagidanticDict((1, 2, 3))


class TestPaginatorSetProxy:
    """Tests for Test PagidanticSet"""

    def test_init(self, paginator_set_proxy: PagidanticSet):
        """Validate init assigment."""
        assert paginator_set_proxy.object_list == [1, 2, 3, 4, 5, 6, 7, 8]
        assert paginator_set_proxy.page_limit == 3
        assert paginator_set_proxy.start_page == 0

    def test_invalid_object_list(self):
        """Validate object_list param."""
        with pytest.raises(TypeError):
            PagidanticSet([1, 2, 3])
        with pytest.raises(TypeError):
            PagidanticSet({"a": 1})
        with pytest.raises(TypeError):
            PagidanticSet((1, 2, 3))



from pagidantic import PagidanticDict, PagidanticFactory, PagidanticSet, Paginator


def get_paginator(object_list):
    factory = PagidanticFactory(type(object_list).__name__)
    return factory.get_paginator()


class TestPaginatorFactory:
    """Tests for PaginatorFactory."""

    def test_factory_result(self):
        """
        Test if expected paginator class/subclass is returned based on object_list input type.
        """
        object_list = [1, 2]
        assert get_paginator(object_list) == Paginator
        object_list = (1, 2, 3)
        assert get_paginator(object_list) == Paginator
        object_list = {"a": 1, "b": 3}
        assert get_paginator(object_list) == PagidanticDict
        object_list = {1, "b", True}
        assert get_paginator(object_list) == PagidanticSet



import pytest
from pydantic import ValidationError

from pagidantic import Paginator, pagidantic


class TestDefaultPaginator:
    """Tests for Paginator object."""

    def test_pagidantic_init(self):
        """Test pagidantic function initial values."""
        assert pagidantic(object_list=[], page_limit=5, start_page=0)
        assert pagidantic(object_list={1, "a"})
        assert pagidantic(object_list=(1, "a"))
        assert pagidantic(object_list={"a": 1})
        assert pagidantic(object_list=[], page_limit=True, start_page=0)
        assert pagidantic(object_list=[], page_limit=True, start_page=False)
        assert pagidantic(object_list=[], page_limit=False, start_page="0")
        with pytest.raises(TypeError):
            assert pagidantic(object_list=None, page_limit=5, start_page=0)
        with pytest.raises(TypeError):
            assert pagidantic(object_list="", page_limit=5, start_page=0)
        with pytest.raises(TypeError):
            assert pagidantic(object_list=0, page_limit=5, start_page=0)
        with pytest.raises(TypeError):
            assert pagidantic(object_list=True, page_limit=5, start_page=0)
        with pytest.raises(ValidationError):
            assert pagidantic(object_list=[], page_limit="str", start_page=0)
        with pytest.raises(ValidationError):
            assert pagidantic(object_list=[], page_limit=3, start_page="str")
        with pytest.raises(ValidationError):
            assert pagidantic(object_list=[], page_limit=None, start_page=0)
        with pytest.raises(ValidationError):
            assert pagidantic(object_list=[], page_limit=None, start_page=None)
        with pytest.raises(ValidationError):
            assert pagidantic(object_list=[], page_limit=3, start_page=None)

    def test_init(self, paginator: Paginator, list_of_int: list[int]):
        """Validate init assigment."""
        assert paginator.object_list == list_of_int
        assert paginator.page_limit == 10
        assert paginator.start_page == 0

    def test_init_valid_object_list(self):
        """Validate object list."""
        paginator_list = pagidantic([1, 2, 3])
        assert paginator_list.object_list == [1, 2, 3]
        paginator_tuple = pagidantic((1, 2, 3))
        assert paginator_tuple.object_list == [1, 2, 3]
        paginator_dict = pagidantic({"a": 1, "b": 2})
        assert paginator_dict.object_list == [
            {
                "a": 1,
            },
            {"b": 2},
        ]
        paginator_set = pagidantic({1, 2, 3})
        assert paginator_set.object_list == [1, 2, 3]

    def test_response(self, paginator: Paginator):
        """Validate current object data."""
        expected = {
            "total_pages": 4,
            "data": list(range(1, 11)),
            "page_number": 0,
            "has_next": True,
            "has_previous": False,
        }
        assert paginator.response == expected

    def test_get_next(self, paginator: Paginator):
        """Validate if response data is changed after calling next page data."""
        paginator.get_next()
        expected = {
            "total_pages": 4,
            "data": list(range(11, 21)),
            "page_number": 1,
            "has_next": True,
            "has_previous": True,
        }
        assert paginator.response == expected

    def test_get_previous(self):
        """Validate if response data is changed after calling next previous data."""
        p = pagidantic(object_list=[1, 2, 3, 4, 5], page_limit=2, start_page=1)
        p.get_previous()
        expected = {
            "total_pages": 2,
            "data": list(range(1, 3)),
            "page_number": 0,
            "has_next": True,
            "has_previous": False,
        }
        assert p.response == expected

    def test_get_current_data(self, paginator: Paginator):
        """
        Validate if page\'s object list is changed after multiple next/previous page calls.
        """
        paginator.get_next()
        paginator.get_next()
        assert paginator.page.object_list == list(range(21, 31))
        paginator.get_previous()
        assert paginator.page.object_list == list(range(11, 21))
        paginator.get_previous()
        assert paginator.page.object_list == list(range(1, 11))
        paginator.get_previous()
        paginator.get_previous()
        paginator.get_previous()
        paginator.get_previous()
        assert paginator.page.object_list == list(range(1, 11))

    def test_get_previous_from_first_page(self, paginator: Paginator):
        """
        Check if response data is unchanged after previous page call while start_page is set to 0.
        """
        paginator.get_previous()
        expected = {
            "total_pages": 4,
            "data": list(range(1, 11)),
            "page_number": 0,
            "has_next": True,
            "has_previous": False,
        }
        assert paginator.response == expected

    def test_get_next_previous(self, paginator: Paginator):
        """Check if response data is changed correctly after multiple next/previous page calls."""
        paginator.get_next()
        paginator.get_next()
        paginator.get_next()
        expected = {
            "total_pages": 4,
            "data": list(range(31, 41)),
            "page_number": 3,
            "has_next": True,
            "has_previous": True,
        }
        assert paginator.response == expected
        paginator.get_previous()
        expected = {
            "total_pages": 4,
            "data": list(range(21, 31)),
            "page_number": 2,
            "has_next": True,
            "has_previous": True,
        }
        assert paginator.response == expected

    def test_get_page_response(self, paginator: Paginator):
        """Validate if requested page data is correct."""
        response = paginator.get_page_response(4)
        expected = {
            "total_pages": 4,
            "data": list(range(41, 51)),
            "page_number": 4,
            "has_next": False,
            "has_previous": True,
        }
        assert response == expected

    def test_get_page_response_page_number(self, paginator: Paginator):
        """Validate page_number param in get_page_response method."""
        assert paginator.get_page_response(1)
        assert paginator.get_page_response(True)
        assert paginator.get_page_response(False)
        with pytest.raises(TypeError):
            assert paginator.get_page_response("test")
        with pytest.raises(TypeError):
            assert paginator.get_page_response("1")

    def test_total(self, paginator: Paginator):
        """Validate total property."""
        assert paginator.total == 50

    def test_total_pages(self, paginator: Paginator):
        """Validate total_pages property."""
        assert paginator.total_pages == 4

    def test_page_range(self, paginator: Paginator):
        """Validate page_range property."""
        assert paginator.page_range == range(5)

    def test_get_objects(self, paginator: Paginator):
        """Validate get_objects method with given page number will return expected object_list."""
        assert paginator.get_objects(1) == list(range(11, 21))

    def test_get_objects_page_number(self, paginator: Paginator):
        """Validate page_number param in get_objects method."""
        assert paginator.get_objects(1)
        assert paginator.get_objects(True)
        assert paginator.get_objects(False)
        with pytest.raises(TypeError):
            assert paginator.get_objects("test")
        with pytest.raises(TypeError):
            assert paginator.get_objects("1")

    def test_has_next(self, paginator: Paginator):
        """Check if paginator have next page."""
        assert paginator.has_next

    def test_not_has_next(self):
        """Check if paginator does not have next page."""
        p = pagidantic(object_list=[1, 2, 3], page_limit=5)
        assert not p.has_next

    def test_has_previous(self):
        """Check if paginator have previous page."""
        p = pagidantic(object_list=[1, 2, 3, 4, 5], page_limit=2, start_page=1)
        assert p.has_previous

    def test_not_has_previous(self, paginator: Paginator):
        """Check if paginator does not have previous page."""
        assert not paginator.has_previous

    def test_iter(self, paginator: Paginator):
        """Check __iter__ method."""
        for current_page in paginator:
            expected = {
                "total_pages": 4,
                "data": current_page.object_list,
                "page_number": current_page.page_number,
                "has_next": current_page.has_next(),
                "has_previous": current_page.has_previous(),
            }
            assert paginator.response == expected

    def test_page(self, paginator: Paginator):
        """Check if Page creates expected data."""
        page = paginator._page(
            object_list=list(range(1, 11)), page_number=0, paginator=paginator
        )
        assert page.object_list == list(range(1, 11))
        assert page.has_next()
        assert not page.has_previous()
        assert page.count == 10

    def test_page_count(self):
        """Validate Page 'count' with different start_page value while iterating over paginator."""
        object_list = [1, 2, 3, 4, 5]
        limit = 2
        paginator = pagidantic(object_list=object_list, page_limit=limit)
        for index, page in enumerate(paginator):
            if index == 2:  # last page in this example
                assert page.count == 1
            else:
                assert page.count == 2

        paginator = pagidantic(object_list=object_list, page_limit=limit, start_page=1)
        for index, page in enumerate(paginator):
            if index == 1:  # last page in this example
                assert page.count == 1
            else:
                assert page.count == 2
        paginator = pagidantic(object_list=object_list, page_limit=limit, start_page=2)
        for page in paginator:
            assert page.count == 1



from pagidantic import Paginator


class TestListOfDictPaginator:
    """Tests for paginator object with list of dict as input."""

    def test_init(self, second_paginator: Paginator, list_of_dict: list[dict]):
        assert second_paginator.object_list == list_of_dict
        assert second_paginator.page_limit == 3
        assert second_paginator.start_page == 1

    def test_response(self, second_paginator: Paginator):
        expected = {
            "total_pages": 6,
            "data": [{"4": 4}, {"5": 5}, {"6": 6}],
            "page_number": 1,
            "has_next": True,
            "has_previous": True,
        }
        assert second_paginator.response == expected



from typing import Type, Union

from pagidantic.factory import PagidanticFactory
from pagidantic.paginator import Paginator


def pagidantic(
    object_list: Union[list, tuple, dict, set],
    page_limit: int = 10,
    start_page: int = 0,
):
    """
    This function create a paginator instance based on the type of the object list passed

    :param object_list: list, tuple, dict or set of objects to be paginated.
    :type object_list: Union[list, tuple, dict, set]
    :param page_limit: Number of items per page.
    :type page_limit: int, optional
    :param start_page: The page to start pagination from.
    :type start_page: int, optional
    :return: paginator instance
    :rtype: Paginator
    """
    factory = PagidanticFactory(type(object_list).__name__)
    paginator: Type[Paginator] = factory.get_paginator()
    return paginator(
        object_list=object_list,
        page_limit=page_limit,
        start_page=start_page,
    )



import math
from functools import cached_property
from typing import Sequence, Union

from pydantic.dataclasses import dataclass

from pagidantic.page import Page


@dataclass
class Paginator:
    """
    Objects paginator. Should be initialised using paginate function.

    Arguments:
        :object_list: list of sequential objects that will be split across pages.
        :page_limit: number of objects per page.
        :start_page: number of page from which paginator will start.
    """

    object_list: Union[list, tuple, dict, set]
    page_limit: int = 10
    start_page: int = 0

    def __post_init_post_parse__(self):
        """Executed after initial validation. Set initial page."""
        self.page: Page = self._page(
            object_list=self.get_objects(self.start_page),
            page_number=self.start_page,
            paginator=self,
        )

    def __iter__(self):
        """Iterate over paginator pages. Every iteration updates paginator page object."""
        for _ in self.page_range:
            yield self.page
            self.get_next()

    def get_objects(self, page_number: int) -> Sequence:
        """Retrieve page list of data."""
        if not isinstance(page_number, int):
            raise TypeError(f"{page_number} expected to be int.")
        n = self.page_limit * page_number
        return self.object_list[n : n + self.page_limit]

    @property
    def response(self):
        """Retrieve response result property."""
        data = {
            "data": self.page.object_list,
            "page_number": self.page.page_number,
            "has_next": self.has_next,
            "has_previous": self.has_previous,
        }
        return self._create_response(**data)

    @property
    def has_next(self) -> bool:
        """Page's "has next" method."""
        return self.page.has_next()

    @property
    def has_previous(self) -> bool:
        """Page's "has previous" method."""
        return self.page.has_previous()

    def get_next(self) -> None:
        """Get next page. Overrides paginator\'s page attribute"""
        if self.has_next:
            self.page.page_number += 1
            next_page = self._page(
                object_list=self.get_objects(self.page.page_number),
                page_number=self.page.page_number,
                paginator=self,
            )
            self.page = next_page  # noqa

    def get_previous(self) -> None:
        """Get previous page. Overrides paginator\'s page attribute."""
        if self.has_previous:
            self.page.page_number -= 1
            previous_page = self._page(
                object_list=self.get_objects(self.page.page_number),
                page_number=self.page.page_number,
                paginator=self,
            )
            self.page = previous_page  # noqa

    @staticmethod
    def _page(*args, **kwargs) -> Page:
        """Returns Page object."""
        return Page(*args, **kwargs)

    def get_page_response(self, page_number: int = 0):
        """
        Get response of requested page number.
        number=0 equals first page.
        """
        if not isinstance(page_number, int):
            raise TypeError(f"{page_number} expected to be int.")
        page = self._page(
            object_list=self.get_objects(page_number),
            page_number=page_number,
            paginator=self,
        )
        data = {
            "data": page.object_list,
            "page_number": page_number,
            "has_next": page.has_next(),
            "has_previous": page.has_previous(),
        }
        return self._create_response(**data)

    @cached_property
    def total(self):
        """Return the total number of objects, across all pages."""
        return len(self.object_list)

    @property
    def total_pages(self):
        """Number of total pages. Lack of additional pages means total is 0."""
        return 0 if self.total == 0 else math.ceil(self.total / self.page_limit) - 1

    @property
    def page_range(self):
        """Return a range of pages."""
        return range(self.total_pages + 1 - self.start_page)

    def _create_response(self, **kwargs):
        """Creates json response object."""
        return {
            "total_pages": self.total_pages,
            "data": kwargs["data"],
            "page_number": kwargs["page_number"],
            "has_next": kwargs["has_next"],
            "has_previous": kwargs["has_previous"],
        }



from pagidantic.paginator import Paginator


class PagidanticSet(Paginator):
    """
    Initialize a Paginator for paginating a set.

    :param object_list: The set to be paginated.
    :type object_list: dict
    :param page_limit: Number of items per page.
    :type page_limit: int, optional
    :param start_page: The page to start pagination from.
    :type start_page: int, optional
    """

    def __init__(self, object_list: set, page_limit: int = 10, start_page: int = 0):
        self.object_list = object_list
        super().__init__(
            object_list=self._set_to_list(),
            page_limit=page_limit,
            start_page=start_page,
        )

    def _set_to_list(self) -> list[set]:
        """Transform set to list of sets."""
        if not isinstance(self.object_list, set):
            raise TypeError(f"Expected set object, not {type(self.object_list)}")
        return list(self.object_list)



from typing import Sequence


class Page:
    """
    A Page object, which contains a list of objects and some additional information such as the current page number and
    the paginator instance.
    """

    def __init__(self, object_list: Sequence, page_number: int, paginator):
        """
        Initialize a Page object with a list of objects, the current page number, and a paginator instance.

        :param object_list: A list of objects to be paginated.
        :type object_list: Sequence
        :param page_number: The current page number.
        :type page_number: int
        :param paginator: The paginator instance.
        :type paginator: object
        """
        self.object_list = object_list
        self.page_number = page_number
        self.paginator = paginator

    def has_next(self) -> bool:
        """
        Check if paginator has next page.

        :return: True if the paginator has a next page, False otherwise.
        :rtype: bool
        """
        return self.page_number < self.paginator.total_pages

    def has_previous(self) -> bool:
        """
        Check if paginator has previous page.

        :return: True if the paginator has a previous page, False otherwise.
        :rtype: bool
        """
        return self.page_number > 0

    @property
    def count(self):
        """
        Return a number of page objects.

        :return: The number of objects in the current page.
        :rtype: int
        """
        return len(self.object_list)



"""
Pagidantic is a Python package for pagination using Pydantic.

It's easy to use, lightweight, and easy to integrate with existing projects.

It helps in creating efficient pagination solution with minimal code, while maintaining type checking and data validation with Pydantic.
"""


__version__ = "1.0.0"

from pagidantic.dict import PagidanticDict
from pagidantic.factory import PagidanticFactory
from pagidantic.page import Page
from pagidantic.pagidantic import pagidantic
from pagidantic.paginator import Paginator
from pagidantic.set import PagidanticSet

__all__ = [
    "PagidanticFactory",
    "Paginator",
    "pagidantic",
    "PagidanticDict",
    "PagidanticSet",
    "Page",
]



from typing import Type

from pydantic.dataclasses import dataclass

from pagidantic.dict import PagidanticDict
from pagidantic.paginator import Paginator
from pagidantic.set import PagidanticSet


@dataclass
class PagidanticFactory:
    """A factory class for creating paginator instances based on the type of objects to be paginated."""

    objects_type: str

    def get_paginator(self) -> Type[Paginator]:
        """
        Returns paginator class/subclass.

        :return: The appropriate paginator class/subclass based on the type of objects to be paginated.
        :rtype: Type[Paginator]
        :raise TypeError: If object_list is of an unsupported type.
        """
        if self.objects_type in ["list", "tuple"]:
            return Paginator
        elif self.objects_type == "dict":
            return PagidanticDict
        elif self.objects_type == "set":
            return PagidanticSet
        else:
            raise TypeError(
                f"Unsupported type {self.objects_type} for object_list param."
            )
